#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include "libhtmlparse.h"


int commentstart(){
	return 0;
}

int commentend(){
	return 0;
}

int comments (char *comment) {
        printf("Comment detected: %s\n", comment);
        return 0;
}

int entity(char *text){
        printf("Entity detected: %s\n", text);
        return 0;
}

int textstart(){
	return 0;
}

int textend(){
	return 0;
}

int texts (char *comment) {
        printf("Text detected: %s\n", comment);
        return 0;
}



int XHTMLtag (char *tag, struct ArgvTable *args, int numargs){

	int i;

	printf("XHTML tag %s detected\n", tag);

	if (args) {
		printf("\twith %d args:\n", numargs);
		for(i=0; i < numargs; i++) {
			printf("\t\t%s=%s\n", args[i].arg, args[i].val);
		}
	}
	return 0;
}

int phpcode (char *comment) {
        printf("PHP detected: %s\n", comment);
        return 0;
}

int xmltag (char *tag, struct ArgvTable *args, int numargs){

	int i;

	printf("XML tag detected\n");

	if (args) {
		printf("\twith %d args:\n", numargs);
		for(i=0; i < numargs; i++) {
			printf("\t\t%s=%s\n", args[i].arg, args[i].val);
		}
	}
	return 0;

}

int endtag (char *tag) {
        printf("End tag %s detected\n", tag);
        return 0;
}

int decltag (char *tag, struct ArgvTable *args, int numargs) {
        int i;

        printf("Declaration tag %s detected\n", tag);
        if (args) {
                printf("\twith %d args:\n", numargs);
                for(i=0; i < numargs; i++) {
                        printf("\t\t%s=%s\n", args[i].arg, args[i].val);
                }
        }
        return 0;
}

int starttag (char *tag, struct ArgvTable *args, int numargs) {
	int i;

        printf("Start tag %s detected\n", tag);
	if (args) {
		printf("\twith %d args:\n", numargs);
		for(i=0; i < numargs; i++) {
			printf("\t\t%s=%s\n", args[i].arg, args[i].val);
		}
	}
        return 0;
}



void open_error () {
        perror("open");
        exit(EXIT_FAILURE);
}

void usage (char **argv) {
        fprintf(stderr, "Usage: %s [xpctesdXy][HTML file]\n"
			"	x = include xmltest     \n"
			"	p = include phptest     \n"
			"	c = include comments    \n"
			"       t = include text        \n"
			"       e = include endtags     \n"
			"       s = include starttags   \n"
			"       d = include declaration \n"
			"	X = include XHTML       \n"
			"	y = include entities    \n"
			"			        \n"
			"You may mix up your own drink. \n", argv[0]);
        exit(0);
}

int main (int argc, char **argv) {
        FILE *fp;
	char *html = NULL;
	clock_t filereads, filereade, parses, parsee;
	struct stat filer;

        if (argc < 3) {
                usage(argv);
        }

	filereads = clock();

        fp = fopen(argv[2], "r");

        if (!fp) {
                open_error();
        }

	stat(argv[2], &filer);

	html = calloc(filer.st_size + 1, 1);

	if (!html) {
		printf("memory error\n");
		exit(EXIT_FAILURE);
	}

	fread(html, 1, filer.st_size, fp);
        fclose(fp);

        filereade = clock();

if(strstr(argv[1],"c")){
        commentCallBack		= comments;
	commentStartCallBack	= commentstart;
	commentEndCallBack	= commentend;
}
if(strstr(argv[1],"t")){
	textCallBack		= texts;
	textStartCallBack	= textstart;
	textEndCallBack		= textend;
}
if(strstr(argv[1],"e")){
	endCallBack		= endtag;
}
if(strstr(argv[1],"s")){
	startCallBack		= starttag;
}
if(strstr(argv[1],"d")){
	declCallBack		= decltag;
}
if(strstr(argv[1],"p")){
	phpCallBack		= phpcode;
}
if(strstr(argv[1],"X")){
	XHTMLCallBack		= XHTMLtag;
}

if(strstr(argv[1],"x")){
	xmlCallBack		= xmltag;
}

if(strstr(argv[1],"y")){
	entityCallBack		= entity;
}
	parses = clock();

        (void)parse(html);

	parsee = clock();
	free(html);

	printf("\n=================================================================\n");
	printf("If you see this message, that mean that libhtmlparse successfully\n");
	printf("parsed your file without complaint. There should be an outline of\n");
	printf("the HTML above this text.\n");
	printf("=================================================================\n");
	printf("Benchmarks:\n\nFile Read: %0.3f second(s)\nHTML Parse: %0.3f second(s)\n",
		(((double)filereade - filereads) / CLOCKS_PER_SEC),
		(((double)parsee - parses)) / CLOCKS_PER_SEC);
	printf("=================================================================\n");
        return 0;
}
