/* libhtmlparse: Quick and simple HTML parser
   Mooneer Salem (mooneer@translator.cx) */

#include "libhtmlparse.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

/* needed to pass text in <script> tags verbatim */
static int lhtml_script_passthru;

#ifdef _INLINE_CODE
#ifndef inline
#define inline __inline__ /* works also with -ansi */
#endif
#endif

static int numArgs=0, numArgsStatus=0;

/* Define INLINE here, if you mean, that the result is faster
   May depend on your compiler  */

#ifdef _INLINE_CODE
inline
#endif

/* argument caching (e.g width="80%") */

struct ArgvTable *addArgToTable (struct ArgvTable *args, char *arg, char *val) {
        numArgs++;
        if (!args) {
                args = (struct ArgvTable*) calloc(1, sizeof(struct ArgvTable)*(numArgs+1));
        } else args = (struct ArgvTable*) realloc(args, sizeof(struct ArgvTable)*(numArgs+1));

        if (!args) {
                printf("memory error\n");
                exit(EXIT_FAILURE);
        }

        args[numArgs-1].arg = arg;
        args[numArgs-1].val = val;
        return args;
}


#ifdef _INLINE_CODE
inline
#endif

/* clean up memory */

void freeArgs (struct ArgvTable *args) {
        register int i;

        if(args) {
                for(i=0; i<numArgs; i++) {
                        free(args[i].arg);
                        free(args[i].val);
                }

                free(args);
                args=NULL;
                numArgs=0;
        }
}


#ifdef _INLINE_CODE
inline
#endif
const char *parseText(const char *html) {
        char *tmp;
        const char *tmp2;
        register int ret=0;

        while(*html && isspace(*html)) html++;

        if (*html == '<') return html;

        tmp2 = html;
        while(*html && *html != '<') html++;

        tmp = (char *)calloc(1, html-tmp2+1);
        if (!tmp) return "";

        memcpy(tmp, tmp2, html-tmp2);

	if (strlen(tmp) > 0) {
	        if (textStartCallBack) {
        	        ret = (*textStartCallBack)();
                	if (ret) {
                        	free(tmp);
                	        return "";
        	        }
	        }

	        if (textCallBack) {

			if (entityCallBack){ /* that is textCallBack(text)
						with entityCallBack(entity) as an extrabonus */
				//printf("entity is here\n");
				parseForEntities(tmp);
			}

			else{
        	        	ret = (*textCallBack)(tmp);
                		if (ret) {
                        		free(tmp);
                 	       		return "";
                			}
			}
        	}

	        if (textEndCallBack) {
        	        ret = (*textEndCallBack)();
                	if (ret) {
                        	free(tmp);
                 	       return "";
                	}
        	}
	}

        free(tmp);

        if (*(html+1) == '>') html += 2;
        return html;
}


#ifdef _INLINE_CODE
inline
#endif
const char *parseComment (const char *html) {
        char *tmp;
        const char *tmp2;
        register int ret=0;


	while(*html == '-' || isspace(*html))html++;

        tmp2 = html;
        while(*html && !(*html == '-' && *(html+1) == '-' && *(html+2) == '>')) html++;

        tmp = (char *)calloc(1, html-tmp2+1);
        if (!tmp) return "";

        memcpy(tmp, tmp2, html-tmp2);

	if (*(html+3)) html += 3;

        if (commentStartCallBack) {
                ret = (*commentStartCallBack)();
                if (ret) {
                        free(tmp);
                        return "";
                }
        }

        if (commentCallBack) {
                ret = (*commentCallBack)(tmp);
                if (ret) {
                        free(tmp);
                        return "";
                }
        }

        if (commentEndCallBack) {
                ret = (*commentEndCallBack)();
                if (ret) {
                        free(tmp);
                        return "";
                }
        }

        free(tmp);
        return html;
}



#ifdef _INLINE_CODE
inline
#endif
const char *parseEndTag(const char *html) {
        char *tmp;
        const char *tmp2;
        register int ret=0;

        html++;

        tmp2 = html;
        while(*html && *html != '>') html++;

        tmp =(char *) calloc(1, html-tmp2+1);
        if (!tmp) return "";

        memcpy(tmp, tmp2, html-tmp2);

//        if (endCallBack) {
                ret = (*endCallBack)(tmp);
                if (ret) {
                        free(tmp);
                        return "";
                }
//        }

        if (*html == '>') html++;

        free(tmp);
        return html;
}


#ifdef _INLINE_CODE
inline
#endif
const char *parsePHP(const char *html) {
	const char *tmp;
	char *tmp2;
	register int ret=0;

	html += 4;
	while(*html && isspace(*html)) html++;

	tmp = html;

	while(*html && !(*html == '?' && *(html+1) == '>')) html++;
	tmp2 = (char *)calloc(1, html-tmp+1);
	if (!tmp2) return "";

	memcpy(tmp2, tmp, html-tmp);

//	if (phpCallBack) {
		ret = (*phpCallBack)(tmp2);
		if (ret) {
			free(tmp2);
			return "";
		}
//	}

	free(tmp2);
	html += 2;
	return html;

}



#ifdef _INLINE_CODE
inline
#endif

/* parse the XML tag itself */

const char *parseXMLtag(const char *html) {

        char *tag, *name, *value;
        const char *tmp;
        int ret;
        struct ArgvTable *tmp2 = NULL;

        numArgs = 0;
        tmp = html;
        while(*html && !isspace(*html) && *html != '>') html++;

	/* you may want to upper/lower tags, so I leave the tag itself untouched */
        tag = (char *)calloc(1, html-tmp+1);
        if (!tag) {
                return "";
        }

        memcpy(tag, tmp, html-tmp);

        if (*html == '>') {

                ret = (*xmlCallBack)(tag, NULL, 0);
                free(tag);
                if (*html == '>') html++;
                return((ret) ? (char *) "" : html);

        }


        while(*html && isspace(*html)) html++;

        while(*html && *html != '>' ) {
               while(isspace(*html)) html++;
                if (*html == '>') break;


                tmp = html;
                while(*html && !isspace(*html) && *html != '=' && *html != '>') html++;
                name = (char *)calloc(1, html-tmp+1);
                if (!name) {
                        free(tag);
                        return "";
                }

                memcpy(name, tmp, html-tmp);

                if (isspace(*html)) {
                        tmp2 = addArgToTable(tmp2, name, NULL);
                        while(*html && isspace(*html) && *html != '>') html++;
                }


                if (*html == '>') {
                        tmp2 = addArgToTable(tmp2, name, NULL);
                        html++;
                        break;
                }

                if (*html == '=') html++;

                if (*html != '"' && *html != '\'') {
//                        html++;
                        tmp = html;
                        while(*html && *html != '>' && !isspace(*html)) html++;
                        value = (char *)calloc(1, html-tmp+1);
                        if (!value) {
                                free(name);
                                free(tag);
                                freeArgs(tmp2);
                                return "";
                        }

                        memcpy(value, tmp, html-tmp);
                        tmp2 = addArgToTable(tmp2, name, value);
                } else if (*html == '"') {
                        html++;
                        tmp = html;
                        while(*html && !(*html == '"' && *(html-1) != '\\')) html++;
                        value = (char *) calloc(1, html-tmp+1);
                        if (!value) {
                                free(name);
                                free(tag);
                                freeArgs(tmp2);
                                return "";
                        }

                        memcpy(value, tmp, html-tmp);
                        html++;
                        tmp2 = addArgToTable(tmp2, name, value);
                } else if (*html == '\'') {
                        html++;
                        tmp = html;
                        while(*html && !(*html == '\'' && *(html-1) != '\\')) html++;

                        value =  (char *)calloc(1, html-tmp+1);
                        if (!value) {
                                free(name);
                                free(tag);
                                freeArgs(tmp2);
                                return "";
                        }

                        memcpy(value, tmp, html-tmp);
                        html++;
                        tmp2 = addArgToTable(tmp2, name, value);
		}

                tmp = NULL;
        }

        if (*html) html++;

        ret = (*xmlCallBack)(tag, tmp2, numArgs);



	if(tmp2){
		freeArgs(tmp2);
        }

	free(tag);

        numArgsStatus=0;

        return (ret ? "" : html);

}

#ifdef _INLINE_CODE
inline
#endif

/* cannibalistic function, munches the actuall tag */

const char *eatUp(const char *html){

	while(*html != '>'){
		html++;
	}
	html++;
	return html;
}

#ifdef _INLINE_CODE
inline
#endif

/* cannibalistic function, munches the actuall text */

const char *eatUpText(const char *html){

	while(*html && *html != '<'){
		html++;
	}

	return html;
}


#ifdef _INLINE_CODE
inline
#endif

/* decides, if a found '?' leads to PHP or XML if requisited
   otherwise it gormandizes them up. *burps* */

const char *parseXML(const char *html) {

	/* conditional expressions inside a conditional expression
	   don't try _this_ at home kids! ;-) */

	 html=(((tolower(*(html+1)))=='p') ?
	( (phpCallBack) ? parsePHP   (html) :  eatUp(html) ) :
	( (xmlCallBack) ? parseXMLtag(html) :  eatUp(html) )   );

	 return html;
}


#ifdef _INLINE_CODE
inline
#endif
const char *parseStartTag (const char *html) {
        char *tag, *name, *value;
        const char *tmp;
        int ret;
        struct ArgvTable *tmp2 = NULL;

        numArgs = 0;
        tmp = html;
        while(*html && !isspace(*html) && *html != '>' && *html != '/') html++;

        tag = (char *)calloc(1, html-tmp+1);
        if (!tag) {
                return "";
        }

        memcpy(tag, tmp, html-tmp);

	if (!strcmp("script", tag)) {
		lhtml_script_passthru = 1;
	}

        if (*html == '>') {
                if (startCallBack) {
                        ret = (*startCallBack)(tag, NULL, 0);
                        free(tag);

                        if (*html == '>') html++;
                	return((ret) ? "" : html);
                }
        }
	else if (*html == '/' ) {   /* XHTML empty tag like <hr/>, <br/>*/

		/**********************************************
		* You may choose now between two behaviors    *
		* of libhtmlparse to handle XHTML empty tags: *
		* a) call XHTMLCallBack                       *
		* b) call start- AND endCallBack              *
		***********************************************/


		if (startCallBack && !(XHTMLCallBack)) {
			ret = (*startCallBack)(tag, NULL, 0);
		}
		if (endCallBack && !ret && !(XHTMLCallBack)) {
			ret = (*endCallBack)(tag);
		}
		if(XHTMLCallBack){
			ret = (*XHTMLCallBack)(tag, NULL, 0);
		}


		free(tag);
		html += 2;
		return((ret) ? "" : html);
	}

        while(*html && isspace(*html)) html++;

        while(*html && *html != '>' ) {
               while(isspace(*html)) html++;
                if (*html == '>') break;

		if (*html == '/' && *(html+1) == '>') {
			html++; break;
		}

                tmp = html;
                while(*html && !isspace(*html) && *html != '=' && *html != '>') html++;
                name = (char *)calloc(1, html-tmp+1);
                if (!name) {
                        free(tag);
                        return "";
                }

                memcpy(name, tmp, html-tmp);

                if (isspace(*html)) {
                        tmp2 = addArgToTable(tmp2, name, NULL);
                        while(*html && isspace(*html) && *html != '>' && !(*html == '/' && *(html+1) == '>')) html++;
                }

		if (*html == '/') {
			html++;
			break;
		}

                if (*html == '>') {
                        tmp2 = addArgToTable(tmp2, name, NULL);
                        html++;
                        break;
                }

                if (*html == '=') html++;

                if (*html != '"' && *html != '\'') {
//                        html++;
                        tmp = html;
                        while(*html && *html != '>' && !isspace(*html) && !(*html == '/' && *(html+1) == '>')) html++;
                        value = (char *)calloc(1, html-tmp+1);
                        if (!value) {
                                free(name);
                                free(tag);
                                freeArgs(tmp2);
                                return "";
                        }

                        memcpy(value, tmp, html-tmp);
                        tmp2 = addArgToTable(tmp2, name, value);
                } else if (*html == '"') {
                        html++;
                        tmp = html;
                        while(*html && !(*html == '"' && *(html-1) != '\\')) html++;
                        value = (char *) calloc(1, html-tmp+1);
                        if (!value) {
                                free(name);
                                free(tag);
                                freeArgs(tmp2);
                                return "";
                        }

                        memcpy(value, tmp, html-tmp);
                        html++;
                        tmp2 = addArgToTable(tmp2, name, value);
                } else if (*html == '\'') {
                        html++;
                        tmp = html;
                        while(*html && !(*html == '\'' && *(html-1) != '\\')) html++;

                        value =  (char *)calloc(1, html-tmp+1);
                        if (!value) {
                                free(name);
                                free(tag);
                                freeArgs(tmp2);
                                return "";
                        }

                        memcpy(value, tmp, html-tmp);
                        html++;
                        tmp2 = addArgToTable(tmp2, name, value);
		}

                tmp = NULL;
        }

        if (*html) html++;

        if (startCallBack &&  *(html-2)!='/' ) {
                ret = (*startCallBack)(tag, tmp2, numArgs);
	}
	if (endCallBack && !ret && *(html-2)=='/' && !(XHTMLCallBack)) {
		ret = (*endCallBack)(tag);
	}
	/* these tags may have arguments too, e.g. <hr noshade/> */
	if (XHTMLCallBack && *(html-2)=='/'){
		ret = (*XHTMLCallBack)(tag, tmp2, numArgs);
	}
	if(tmp2){
		freeArgs(tmp2);
	}
	free(tag);

        numArgsStatus=0;

	/* this is a bad hack, feel free to write a better one (maybe a more readable one? ;-)*/
        return
	(XHTMLCallBack) ?
        (html) :
	((ret) ? "" : html);
}


#ifdef _INLINE_CODE
inline
#endif
const char *parseDecl(const char *html) {
        char *tag, *name, *value;
        const char *tmp;
        register int ret=0;
        struct ArgvTable *tmp2 = NULL;

        numArgs = 0;
        tmp = html;
        while(*html && !isspace(*html) && *html != '>') html++;

        tag = (char *)calloc(1, html-tmp+1);
        if (!tag) {
                return "";
        }

        memcpy(tag, tmp, html-tmp);

        if (*html == '>') {
//                if (declCallBack) {
                        ret = (*declCallBack)(tag, NULL, 0);
                        free(tag);

                        if (*html == '>') html++;
                	return((ret) ? "" : html);
                }
//        }

        while(*html && isspace(*html)) html++;

        while(*html && *html != '>') {
                while(isspace(*html)) html++;
                if (*html == '>') break;

                tmp = html;


                switch(*tmp){

                case '\'' :   	html++;
	                	tmp = html;
				while(*html && !(*html == '\'' && *html != '\\')) html++;
			        break;
		case '"'  :     html++;
	                	tmp = html;
				while(*html && !(*html == '"' && *html != '\\')) html++;
                                break;
		default  :
       	        		while(*html && !isspace(*html) && *html != '=' && *html != '>') html++;
				break;
		}

                name = (char *) calloc(1, html-tmp+1);
                if (!name) {
                        free(tag);
                        return "";
                }

                memcpy(name, tmp, html-tmp);

                if (isspace(*html)) {
                        tmp2 = addArgToTable(tmp2, name, NULL);
                        while(*html && isspace(*html) && *html != '>') html++;
			continue;
                }

                if (*html == '>') {
                        tmp2 = addArgToTable(tmp2, name, NULL);
                        html++;
                        break;
                }

		if (*(html+1) == '>') {
			tmp2 = addArgToTable(tmp2, name, NULL);
			html += 2;
			break;
		}

                if (*html == '=') html++;


		switch(*html){

		case '\''  :    html++;
                        	tmp = html;
                        	while(*html && !(*html == '\'' && *(html-1) != '\\')) html++;

                        	value = (char *) calloc(1, html-tmp+1);
                        	if (!value) {
                                	free(name);
                                	free(tag);
                                	freeArgs(tmp2);
                                	return "";
                        	}

                        	memcpy(value, tmp, html-tmp);
                        	html++;
                        	tmp2 = addArgToTable(tmp2, name, value);
                                break;
		case '"'  :     html++;
                        	tmp = html;
                        	while(*html && !(*html == '"' && *(html-1) != '\\')) html++;
                        	value =  (char *)calloc(1, html-tmp+1);
                        	if (!value) {
                        	        free(name);
                        	        free(tag);
                        	        freeArgs(tmp2);
                        	        return "";
                        	}

                        	memcpy(value, tmp, html-tmp);
                        	html++;
                        	tmp2 = addArgToTable(tmp2, name, value);
                                break;
		default  :      html++;
                        	tmp = html;
                        	while(*html && *html != '>' && !isspace(*html)) html++;
                        	value = (char *) calloc(1, html-tmp+1);
                        	if (!value) {
                        	        free(name);
                        	        free(tag);
                        	        freeArgs(tmp2);
                        	        return "";
                        	}

                        	memcpy(value, tmp, html-tmp);
                        	tmp2 = addArgToTable(tmp2, name, value);
			        break;
		}
                tmp = NULL;
        }

        if (*html) html++;

//        if (declCallBack) {
                ret = (*declCallBack)(tag, tmp2, numArgs);
                freeArgs(tmp2);
                return((ret) ? "" : html);
//        }

        freeArgs(tmp2);
        numArgsStatus=0;



        return html;
}


#ifdef _INLINE_CODE
inline
#endif
const char *parseForEntities (const char *tmp){
	char *entity, *text ;
	const char *tmp1, *tmp2;
	register int ret=0;
	while(*tmp){
		tmp1 = tmp;
		while(*tmp && *tmp != '&')tmp++;

		text = (char *)calloc(1, tmp-tmp1+1);
		memcpy(text, tmp1, tmp-tmp1);
		/* the chunk of text before the first entity will
		   not be called, if it starts with an entity*/
		if(strlen(text)>0 && (!(isspace(*text)))){
			if (textCallBack) {
				ret = (*textCallBack)(text);
			}
			free(text);
			tmp1 = "";
		}

		if(*tmp == '&'){
			tmp++;
			tmp2=tmp;
			while(*tmp && *tmp != ';')tmp++;
			entity = (char *)calloc(1, tmp-tmp2+1);
        		if (!entity) {
				/* What is it: an empty entity "&;" or insufficient mem? &%$!
				   Should I add something like if((tmp-tmp2+1) != 0) before
				   allocating entity? */
                			ret = (*entityCallBack)(NULL);
        				}
			else{
				memcpy(entity, tmp2, tmp-tmp2);
				if(*tmp == ';'){  /* should I add an errortrap here? */
					ret = (*entityCallBack)(entity);
					free(entity);
					tmp2 = "";
				}
			}
		}
		if (*tmp) tmp++;
	}

}

struct ErrorTree *parse (const char *html) {
        while(*html) {
              /* while(isspace(*html)){html++;} there may be leading blanks in some autogenerated files
						 add this or not, that is the question ;-)) */

		if (lhtml_script_passthru) {
			const char *text;
			char *tmp;

			text = html;
			while(*text) {
				if (*text == '<') {
					if (*(text+2) == 's' || *(text+2) == 'S') {
						if (*(text+7) == 't' || *(text+7) == 'T') {
							break;
						}
					}
				}
				text++;
			}
			tmp = (char *) malloc(text-html+1);
			strncpy(tmp, html, text-html);

			if (textCallBack) {
				int ret = (*textCallBack)(tmp);
				if (ret) {
					free(tmp);
					return NULL;
				}
			}

			free(tmp);

			lhtml_script_passthru = 0;
			html = text;
		}

                if(*html == '<'){
                        html++;

			switch (*html){
                        case '!'   :
                                html++;

				/* I must admit, I like conditional expressions,
				   they are so obviously obfuscated ;-)          */

				html = (*html == '-') ?
					((commentCallBack) ? parseComment(html) : eatUp(html)) :
					((declCallBack)    ? parseDecl(html)    : eatUp(html))  ;
					break;
                        case '?'  : 			/* XML/PHP tag */
                                	html = (xmlCallBack || phpCallBack) ?
						parseXML(html) :
						eatUp(html);
						break;
                        case '/'  : 			/* HTML end tag */
                                	html = (endCallBack) ?
						parseEndTag(html) :
						eatUp(html);
						break;
                        default  : 			/* HTML start tag */
                                	html = (XHTMLCallBack || startCallBack) ?
						parseStartTag(html) :
						eatUp(html);
						break;
                        }
		      }
                else{				 /* All other text */
                        /* while(isspace(*html))html++;   it seems to be faster inside the function */
                        	html = (textCallBack)  ?
					parseText(html):
					eatUpText(html);
                }
        }

        return NULL;
}

